@@ -3,7 +3,7 @@
  *
  * Home page of code is: http://smartmontools.sourceforge.net
  *
- * Copyright (C) 2004-6 Christian Franke <smartmontools-support@lists.sourceforge.net>
+ * Copyright (C) 2004-12 Christian Franke <smartmontools-support@lists.sourceforge.net>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -11,34 +11,32 @@
  * any later version.
  *
  * You should have received a copy of the GNU General Public License
- * (for example COPYING); if not, write to the Free
- * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * (for example COPYING); If not, see <http://www.gnu.org/licenses/>.
  *
  */
 
+#define WINVER 0x0500
+#define _WIN32_WINNT WINVER
+
+#include "daemon_win32.h"
+
+const char * daemon_win32_cpp_cvsid = "$Id: daemon_win32.cpp 3586 2012-08-06 19:11:20Z chrfranke $"
+  DAEMON_WIN32_H_CVSID;
+
 #include <stdio.h>
 #include <stdlib.h>
 #include <signal.h>
 #include <io.h>
 
 #define WIN32_LEAN_AND_MEAN
-// Need MB_SERVICE_NOTIFICATION (NT4/2000/XP), IsDebuggerPresent() (Win98/ME/NT4/2000/XP)
-#define _WIN32_WINNT 0x0400 
 #include <windows.h>
 #ifdef _DEBUG
 #include <crtdbg.h>
 #endif
 
-#include "daemon_win32.h"
-
-const char *daemon_win32_c_cvsid = "$Id: daemon_win32.cpp,v 1.11 2006/08/09 20:40:20 chrfranke Exp $"
-DAEMON_WIN32_H_CVSID;
-
 
 /////////////////////////////////////////////////////////////////////////////
 
-#define ARGUSED(x) ((void)(x))
-
 // Prevent spawning of child process if debugging
 #ifdef _DEBUG
 #define debugging() IsDebuggerPresent()
@@ -466,79 +464,6 @@ int daemon_detach(const char * ident)
 }
 
 
-/////////////////////////////////////////////////////////////////////////////
-// MessageBox
-
-#ifndef _MT
-//MT runtime not necessary, because mbox_thread uses no unsafe lib functions
-//#error Program must be linked with multithreaded runtime library
-#endif
-
-static LONG mbox_count; // # mbox_thread()s
-static HANDLE mbox_mutex; // Show 1 box at a time (not necessary for service)
-
-typedef struct mbox_args_s {
-	HANDLE taken; const char * title, * text; int mode;
-} mbox_args;
-
-
-// Thread to display one message box
-
-static ULONG WINAPI mbox_thread(LPVOID arg)
-{
-	// Take args
-	mbox_args * mb = (mbox_args *)arg;
-	char title[100]; char text[1000]; int mode;
-	strncpy(title, mb->title, sizeof(title)-1); title[sizeof(title)-1] = 0;
-	strncpy(text , mb->text , sizeof(text )-1); text [sizeof(text )-1] = 0;
-	mode = mb->mode;
-	SetEvent(mb->taken);
-
-	// Show only one box at a time
-	WaitForSingleObject(mbox_mutex, INFINITE);
-	MessageBoxA(NULL, text, title, mode);
-	ReleaseMutex(mbox_mutex);
-
-	InterlockedDecrement(&mbox_count);
-	return 0;
-}
-
-
-// Display a message box
-int daemon_messagebox(int system, const char * title, const char * text)
-{
-	mbox_args mb;
-	HANDLE ht; DWORD tid;
-
-	// Create mutex during first call
-	if (!mbox_mutex)
-		mbox_mutex = CreateMutex(NULL/*!inherit*/, FALSE/*!owned*/, NULL/*unnamed*/);
-
-	// Allow at most 10 threads
-	if (InterlockedIncrement(&mbox_count) > 10) {
-		InterlockedDecrement(&mbox_count);
-		return -1;
-	}
-
-	// Create thread
-	mb.taken = CreateEvent(NULL/*!inherit*/, FALSE, FALSE/*!signaled*/, NULL/*unnamed*/);
-	mb.mode = MB_OK|MB_ICONWARNING
-	         |(svc_mode?MB_SERVICE_NOTIFICATION:0)
-	         |(system?MB_SYSTEMMODAL:MB_APPLMODAL);
-	mb.title = title; mb.text = text;
-	mb.text = text;
-	if (!(ht = CreateThread(NULL, 0, mbox_thread, &mb, 0, &tid)))
-		return -1;
-
-	// Wait for args taken
-	if (WaitForSingleObject(mb.taken, 10000) != WAIT_OBJECT_0)
-		TerminateThread(ht, 0);
-	CloseHandle(mb.taken);
-	CloseHandle(ht);
-	return 0;
-}
-
-
 /////////////////////////////////////////////////////////////////////////////
 
 // Spawn a command and redirect <inpbuf >outbuf
@@ -548,29 +473,13 @@ int daemon_spawn(const char * cmd,
                  const char * inpbuf, int inpsize,
                  char *       outbuf, int outsize )
 {
-	HANDLE pipe_inp_r, pipe_inp_w, pipe_out_r, pipe_out_w, pipe_err_w, h;
-	char temp_path[MAX_PATH];
-	DWORD flags, num_io, exitcode;
-	int use_file, state, i;
-	SECURITY_ATTRIBUTES sa;
-	STARTUPINFO si; PROCESS_INFORMATION pi;
 	HANDLE self = GetCurrentProcess();
 
-	if (GetVersion() & 0x80000000L) {
-		// Win9x/ME: A calling process never receives EOF if output of COMMAND.COM or
-		// any other DOS program is redirected via a pipe. Using a temp file instead.
-		use_file = 1; flags = DETACHED_PROCESS;
-	}
-	else {
-		// NT4/2000/XP: If DETACHED_PROCESS is used, CMD.EXE opens a new console window
-		// for each external command in a redirected .BAT file.
-		// Even (DETACHED_PROCESS|CREATE_NO_WINDOW) does not work.
-		use_file = 0; flags = CREATE_NO_WINDOW;
-	}
-
 	// Create stdin pipe with inheritable read side
+	SECURITY_ATTRIBUTES sa;
 	memset(&sa, 0, sizeof(sa)); sa.nLength = sizeof(sa);
 	sa.bInheritHandle = TRUE;
+	HANDLE pipe_inp_r, pipe_inp_w, h;
 	if (!CreatePipe(&pipe_inp_r, &h, &sa/*inherit*/, inpsize*2+13))
 		return -1;
 	if (!DuplicateHandle(self, h, self, &pipe_inp_w,
@@ -579,36 +488,16 @@ int daemon_spawn(const char * cmd,
 		return -1;
 	}
 
+	// Create stdout pipe with inheritable write side
 	memset(&sa, 0, sizeof(sa)); sa.nLength = sizeof(sa);
 	sa.bInheritHandle = TRUE;
-	if (!use_file) {
-		// Create stdout pipe with inheritable write side
-		if (!CreatePipe(&h, &pipe_out_w, &sa/*inherit*/, outsize)) {
-			CloseHandle(pipe_inp_r); CloseHandle(pipe_inp_w);
-			return -1;
-		}
-	}
-	else {
-		// Create temp file with inheritable write handle
-		char temp_dir[MAX_PATH];
-		if (!GetTempPathA(sizeof(temp_dir), temp_dir))
-			strcpy(temp_dir, ".");
-		if (!GetTempFileNameA(temp_dir, "out"/*prefix*/, 0/*create unique*/, temp_path)) {
-			CloseHandle(pipe_inp_r); CloseHandle(pipe_inp_w);
-			return -1;
-		}
-		if ((h = CreateFileA(temp_path, GENERIC_READ|GENERIC_WRITE,
-			0/*no sharing*/, &sa/*inherit*/, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE) {
-			CloseHandle(pipe_inp_r); CloseHandle(pipe_inp_w);
-			return -1;
-		}
-		if (!DuplicateHandle(self, h, self, &pipe_out_w,
-			GENERIC_WRITE, TRUE/*inherit*/, 0)) {
-			CloseHandle(h); CloseHandle(pipe_inp_r); CloseHandle(pipe_inp_w);
-			return -1;
-		}
+	HANDLE pipe_out_w;
+	if (!CreatePipe(&h, &pipe_out_w, &sa/*inherit*/, outsize)) {
+		CloseHandle(pipe_inp_r); CloseHandle(pipe_inp_w);
+		return -1;
 	}
 
+	HANDLE pipe_out_r;
 	if (!DuplicateHandle(self, h, self, &pipe_out_r,
 		GENERIC_READ, FALSE/*!inherit*/, DUPLICATE_CLOSE_SOURCE)) {
 		CloseHandle(pipe_out_w); CloseHandle(pipe_inp_r); CloseHandle(pipe_inp_w);
@@ -616,6 +505,7 @@ int daemon_spawn(const char * cmd,
 	}
 
 	// Create stderr handle as dup of stdout write side
+	HANDLE pipe_err_w;
 	if (!DuplicateHandle(self, pipe_out_w, self, &pipe_err_w,
 		0, TRUE/*inherit*/, DUPLICATE_SAME_ACCESS)) {
 		CloseHandle(pipe_out_r); CloseHandle(pipe_out_w);
@@ -623,16 +513,18 @@ int daemon_spawn(const char * cmd,
 		return -1;
 	}
 
-	// Create process with pipes/file as stdio
+	// Create process with pipes as stdio
+	STARTUPINFO si;
 	memset(&si, 0, sizeof(si)); si.cb = sizeof(si);
 	si.hStdInput  = pipe_inp_r;
 	si.hStdOutput = pipe_out_w;
 	si.hStdError  = pipe_err_w;
 	si.dwFlags = STARTF_USESTDHANDLES;
+	PROCESS_INFORMATION pi;
 	if (!CreateProcessA(
 		NULL, (char*)cmd,
 		NULL, NULL, TRUE/*inherit*/,
-		flags/*DETACHED_PROCESS or CREATE_NO_WINDOW*/,
+		CREATE_NO_WINDOW, // DETACHED_PROCESS does not work
 		NULL, NULL, &si, &pi)) {
 		CloseHandle(pipe_err_w);
 		CloseHandle(pipe_out_r); CloseHandle(pipe_out_w);
@@ -646,7 +538,9 @@ int daemon_spawn(const char * cmd,
 	CloseHandle(pipe_err_w);
 
 	// Copy inpbuf to stdin
-	// convert \n => \r\n 
+	// convert \n => \r\n
+	DWORD num_io;
+	int i;
 	for (i = 0; i < inpsize; ) {
 		int len = 0;
 		while (i+len < inpsize && inpbuf[i+len] != '\n')
@@ -661,37 +555,25 @@ int daemon_spawn(const char * cmd,
 	}
 	CloseHandle(pipe_inp_w);
 
-	exitcode = 42;
-	for (state = 0; state < 2; state++) {
-		// stdout pipe: read pipe first
-		// stdout file: wait for process first
-		if (state == use_file) {
-			// Copy stdout to output buffer until full, rest to /dev/null
-			// convert \r\n => \n
-			if (use_file)
-				SetFilePointer(pipe_out_r, 0, NULL, FILE_BEGIN);
-			for (i = 0; ; ) {
-				char buf[256];
-				int j;
-				if (!ReadFile(pipe_out_r, buf, sizeof(buf), &num_io, NULL) || num_io == 0)
-					break;
-				for (j = 0; i < outsize-1 && j < (int)num_io; j++) {
-					if (buf[j] != '\r')
-						outbuf[i++] = buf[j];
-				}
-			}
-			outbuf[i] = 0;
-			CloseHandle(pipe_out_r);
-			if (use_file)
-				DeleteFileA(temp_path);
-		}
-		else {
-			// Wait for process exitcode
-			WaitForSingleObject(pi.hProcess, INFINITE);
-			GetExitCodeProcess(pi.hProcess, &exitcode);
-			CloseHandle(pi.hProcess);
+	// Copy stdout to output buffer until full, rest to /dev/null
+	// convert \r\n => \n
+	for (i = 0; ; ) {
+		char buf[256];
+		if (!ReadFile(pipe_out_r, buf, sizeof(buf), &num_io, NULL) || num_io == 0)
+			break;
+		for (int j = 0; i < outsize-1 && j < (int)num_io; j++) {
+			if (buf[j] != '\r')
+				outbuf[i++] = buf[j];
 		}
 	}
+	outbuf[i] = 0;
+	CloseHandle(pipe_out_r);
+
+	// Wait for process exitcode
+	DWORD exitcode = 42;
+	WaitForSingleObject(pi.hProcess, INFINITE);
+	GetExitCodeProcess(pi.hProcess, &exitcode);
+	CloseHandle(pi.hProcess);
 	return exitcode;
 }
 
@@ -841,7 +723,6 @@ static void service_report_status(int state, int seconds)
 {
 	// TODO: Avoid race
 	static DWORD checkpoint = 1;
-	static DWORD accept_more = SERVICE_ACCEPT_PARAMCHANGE; // Win2000/XP
 	svc_status.dwCurrentState = state;
 	svc_status.dwWaitHint = seconds*1000;
 	switch (state) {
@@ -860,17 +741,10 @@ static void service_report_status(int state, int seconds)
 		default:
 			svc_status.dwControlsAccepted =
 				SERVICE_ACCEPT_STOP|SERVICE_ACCEPT_SHUTDOWN|
-				SERVICE_ACCEPT_PAUSE_CONTINUE|accept_more;
+				SERVICE_ACCEPT_PAUSE_CONTINUE|SERVICE_ACCEPT_PARAMCHANGE;
 			break;
 	}
-	if (!SetServiceStatus(svc_handle, &svc_status)) {
-		if (svc_status.dwControlsAccepted & accept_more) {
-			// Retry without SERVICE_ACCEPT_PARAMCHANGE (WinNT4)
-			svc_status.dwControlsAccepted &= ~accept_more;
-			accept_more = 0;
-			SetServiceStatus(svc_handle, &svc_status);
-		}
-	}
+	SetServiceStatus(svc_handle, &svc_status);
 }
 
 
@@ -937,16 +811,15 @@ static char ** svc_main_argv;
 
 // Main function for service, called by service dispatcher
 
-static void WINAPI service_main(DWORD argc, LPSTR * argv)
+static void WINAPI service_main(DWORD /*argc*/, LPSTR * argv)
 {
 	char path[MAX_PATH], *p;
-	ARGUSED(argc);
 
 	// Register control handler
 	svc_handle = RegisterServiceCtrlHandler(argv[0], service_control);
 
 	// Init service status
-	svc_status.dwServiceType = SERVICE_WIN32_OWN_PROCESS|SERVICE_INTERACTIVE_PROCESS;
+	svc_status.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
 	service_report_status(SERVICE_START_PENDING, 10);
 
 	// Service started in \windows\system32, change to .exe directory
@@ -968,26 +841,6 @@ static void WINAPI service_main(DWORD argc, LPSTR * argv)
 /////////////////////////////////////////////////////////////////////////////
 // Windows Service Admin Functions
 
-// Set Service description (Win2000/XP)
-
-static int svcadm_setdesc(SC_HANDLE hs, const char * desc)
-{
-	HINSTANCE hdll;
-	BOOL (WINAPI * ChangeServiceConfig2A_p)(SC_HANDLE, DWORD, LPVOID);
-	BOOL ret;
-	if (!(hdll = LoadLibraryA("ADVAPI32.DLL")))
-		return FALSE;
-	if (!((ChangeServiceConfig2A_p = (BOOL (WINAPI *)(SC_HANDLE, DWORD, LPVOID))GetProcAddress(hdll, "ChangeServiceConfig2A"))))
-		ret = FALSE;
-	else {
-		SERVICE_DESCRIPTIONA sd = { (char *)desc };
-		ret = ChangeServiceConfig2A_p(hs, SERVICE_CONFIG_DESCRIPTION, &sd);
-	}
-	FreeLibrary(hdll);
-	return ret;
-}
-
-
 // Service install/remove commands
 
 static int svcadm_main(const char * ident, const daemon_winsvc_options * svc_opts,
@@ -1016,8 +869,6 @@ static int svcadm_main(const char * ident, const daemon_winsvc_options * svc_opt
 	if (!(hm = OpenSCManager(NULL/*local*/, NULL/*default*/, SC_MANAGER_ALL_ACCESS))) {
 		if ((err = GetLastError()) == ERROR_ACCESS_DENIED)
 			puts(" access to SCManager denied");
-		else if (err == ERROR_CALL_NOT_IMPLEMENTED)
-			puts(" services not implemented on this version of Windows");
 		else
 			printf(" cannot open SCManager, Error=%ld\n", err);
 		return 1;
@@ -1032,13 +883,27 @@ static int svcadm_main(const char * ident, const daemon_winsvc_options * svc_opt
 			CloseServiceHandle(hm);
 			return 1;
 		}
+		// Add quotes if necessary
+		if (strchr(path, ' ')) {
+			i = strlen(path);
+			path[i+1] = '"'; path[i+2] = 0;
+			while (--i >= 0)
+				path[i+1] = path[i];
+			path[0] = '"';
+		}
 		// Append options
 		strcat(path, " "); strcat(path, svc_opts->cmd_opt);
 		for (i = 2; i < argc; i++) {
 			const char * s = argv[i];
-			if (strlen(path)+strlen(s)+1 >= sizeof(path))
+			if (strlen(path)+1+1+strlen(s)+1 >= sizeof(path))
 				break;
-			strcat(path, " "); strcat(path, s);
+			// Add quotes if necessary
+			if (strchr(s, ' ') && !strchr(s, '"')) {
+				strcat(path, " \""); strcat(path, s); strcat(path, "\"");
+			}
+			else {
+				strcat(path, " "); strcat(path, s);
+			}
 		}
 		// Create
 		if (!(hs = CreateService(hm,
@@ -1059,8 +924,10 @@ static int svcadm_main(const char * ident, const daemon_winsvc_options * svc_opt
 			return 1;
 		}
 		// Set optional description
-		if (svc_opts->descript)
-			svcadm_setdesc(hs, svc_opts->descript);
+		if (svc_opts->descript) {
+			SERVICE_DESCRIPTIONA sd = { const_cast<char *>(svc_opts->descript) };
+			ChangeServiceConfig2A(hs, SERVICE_CONFIG_DESCRIPTION, &sd);
+		}
 	}
 	else {
 		// Open
